import sys
import pygame           #load pygame module

class Field():
    def __init__(self, l, x, y):
        self.l = l
        self.x, self.y = x, y
        self.left = self.right = self.top = self.bottom = None
        self.previous = None
        self.isStart = self.isDestination = False
        self.isOpen = self.isClosed = False
        self.isBlocked = False
        self.isWay = False
    
    def toField(self):
        self.draw()
        if self.isStart:
            return "S"
        elif self.isDestination:
            return "D"
        if self.isBlocked:
            return "X"
        elif self.isWay:
            return "."
        if self.isOpen:
            return "o"
        elif self.isClosed:
            return "c"
        return " "
    
    def draw(self):
        if self.isStart:
            color = (50, 255, 255)
        elif self.isDestination:
            color = (255, 255, 50)
        elif self.isBlocked:
            color = (20, 20, 20)
        elif self.isWay:
            color = (100, 100, 100)
        elif self.isOpen:
            color = (50, 255, 50)
        elif self.isClosed:
            color = (200, 200, 200)
        else:
            color = (255, 255, 255)
        self.l.drawField(self.x, self.y, color)

class Labyrinth():
    def __init__(self, dimx, dimy):
        labyrinth = []
        self.start = None
        self.destination = None
        
        base = 500
        self.border = base/50
        self.rectsize = base/dimy
        w = 2*self.border + dimy*self.rectsize
        h = 2*self.border + dimx*self.rectsize
        self.w, self.h = w, h
        self.dimx, self.dimy = dimx, dimy
        
        self.screen = pygame.display.set_mode((w, h))

        pygame.draw.rect(self.screen, (220, 220, 220), (self.border, self.border, h-self.border*2, w-self.border*2))
        
        self.reset()
    
    def reset(self):
        self.labyrinth = [[Field(self, rowid, colid) for colid in range(self.dimy)] for rowid in range(self.dimx)]
        for x, row in enumerate(self.labyrinth):
            for y, f in enumerate(row):
                self.drawField(x, y, (255, 255, 255))
        pygame.display.update()

    def drawField(self, x, y, color):
        pygame.draw.rect(self.screen, color, (self.border+self.rectsize*y+1, self.border+self.rectsize*x+1, self.rectsize-2, self.rectsize-2))
        pygame.display.update()
    
    def translateCoords(self, x, y):
        x, y = x-self.border, y-self.border
        x = x/self.rectsize
        y = y/self.rectsize
        if x < 0 or y < 0: return None
        if x > self.dimy-1: return None
        if y > self.dimx-1: return None
        return y, x #swap
    
    def setStart(self, x, y):
        for row in self.labyrinth:
            for f in row:
                if f.isStart:
                    f.isStart = False
                    f.draw()
        self.start = (x, y)
        self.labyrinth[x][y].isStart = True
        self.labyrinth[x][y].draw()
    
    def setDestination(self, x, y):
        for row in self.labyrinth:
            for f in row:
                if f.isDestination:
                    f.isDestination = False
                    f.draw()
        self.destination = (x, y)
        self.labyrinth[x][y].isDestination = True
        self.labyrinth[x][y].draw()
    
    def blockField(self, x, y):
        self.labyrinth[x][y].isBlocked = True
        self.labyrinth[x][y].draw()
    
    def resetField(self, x, y):
        self.labyrinth[x][y].isBlocked = False
        self.labyrinth[x][y].isDestination = False
        self.labyrinth[x][y].isStart = False
        self.labyrinth[x][y].isOpen = False
        self.labyrinth[x][y].isClosed = False
        self.labyrinth[x][y].isWay = False
        self.labyrinth[x][y].draw()
    
    def show(self): #display map
        tmpMap = self.labyrinth
        
        for row in tmpMap:
            print( " ".join([f.toField() for f in row]) ) #print( row )
        print( "-" * (len(tmpMap[0])*2 - 1) ) #draw line
    
    def prepare(self): #build connections between fields
        for row in self.labyrinth:
            for f in row:
                f.isOpen = False
                f.isClosed = False
                f.isWay = False
                f.draw()
        for x, row in enumerate(self.labyrinth):
            for y, f in enumerate(row):
                if y > 0:
                    f.left = row[y-1]
                if y < len(row)-1:
                    f.right = row[y+1]
                if x > 0:
                    f.top = self.labyrinth[x-1][y]
                if x < len(self.labyrinth)-1:
                    f.bottom = self.labyrinth[x+1][y]
    
    def calc(self):
        self.prepare()
        openNodes = [] #list for next possible nodes
        closedNodes = [] #list for passed nodes
        
        start = self.labyrinth[self.start[0]][self.start[1]]
        destination = self.labyrinth[self.destination[0]][self.destination[1]]
        
        def expand(node): #get next fields
            def add(x):
                if x and not x in openNodes and not x in closedNodes:
                    if x.isBlocked:
                        return
                    x.previous = node #for tracking the way
                    openNodes.append(x) #add to open list
                    x.isOpen = True
                    x.draw()
            add(node.left)
            add(node.right)
            add(node.top)
            add(node.bottom)
        
        def dump():
            self.show()
        
        expand(start) #expand start node
        foundWay = False
        while not foundWay:
            if not openNodes: #no more nodes and we didn't found the destination
                break
            nodes = list(openNodes) #copy current open nodes
            for node in nodes: #process nodes
                if node is destination: #node is destination?
                    foundWay = True
                    break #stop!
                expand(node) #expand node
                openNodes.remove(node) #remove current node from open list
                node.isOpen = False
                node.isClosed = True
                node.draw()
                closedNodes.append(node) #add to closed nodes
            #dump() #show map
        
        if foundWay: #calc way
            way = [(destination.x, destination.y)] #waypoints
            reached = False
            current = destination
            while not reached:
                current = current.previous #walk back
                if not current:
                    break
                current.isWay = True
                current.draw()
                way.insert(0, (current.x, current.y))
                if current is start: #are we there yet?
                    reached = True
        return foundWay
        
if __name__ == "__main__":
    l = Labyrinth(20, 20)
    
    def setStart():
        l.setStart(8, 5)
    
    def setDestination():
        l.setDestination(0, 3)
    
    def setBlocks():
        l.blockField(9, 3); l.blockField(9, 6);
        l.blockField(8, 1); l.blockField(8, 4); l.blockField(8, 8)
        l.blockField(7, 1); l.blockField(7, 2); l.blockField(7, 6); l.blockField(7, 8)
        l.blockField(6, 2); l.blockField(6, 4); l.blockField(6, 6); l.blockField(6, 8)
        l.blockField(5, 0); l.blockField(5, 2); l.blockField(5, 3); l.blockField(5, 4); l.blockField(5, 5); l.blockField(5, 8)
        l.blockField(4, 7)
        l.blockField(3, 1); l.blockField(3, 2); l.blockField(3, 3); l.blockField(3, 5); l.blockField(3, 8); l.blockField(3, 9)
        l.blockField(2, 1); l.blockField(2, 5); l.blockField(2, 7)
        l.blockField(1, 2); l.blockField(1, 3); l.blockField(1, 5); l.blockField(1, 7); l.blockField(1, 8)
        l.blockField(0, 2); l.blockField(0, 5)
    
    mode = 0
    def click():
        x, y = pygame.mouse.get_pos()
        t = l.translateCoords(x, y)
        if t:
            x, y = t
            if mode == 1:
                l.setStart(x, y)
            elif mode == 2:
                l.setDestination(x, y)
            elif mode == 3:
                l.blockField(x, y)
            elif mode == 4:
                l.resetField(x, y)
    
    while True:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                 pygame.quit()
                 sys.exit()
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_c:
                    l.calc()
                elif event.key == pygame.K_s:
                    mode = 1
                elif event.key == pygame.K_d:
                    mode = 2
                elif event.key == pygame.K_b:
                    mode = 3
                elif event.key == pygame.K_n:
                    mode = 4
                elif event.key == pygame.K_f:
                    mode = 0
                elif event.key == pygame.K_a:
                    setStart()
                    setDestination()
                    setBlocks()
                elif event.key == pygame.K_r:
                    l.reset()
                elif event.key == pygame.K_q:
                    pygame.quit()
                    sys.exit()
            elif event.type == pygame.MOUSEBUTTONDOWN:
                click()
